@@ -30,15 +30,17 @@ sf_describe_metadata <- function(verbose=FALSE){
   body_node <- newXMLNode("soapenv:Body", parent=root)  
   body_node <- addChildren(body_node, operation_node)
   
-  if(verbose) {
-    print(base_metadata_url)
-    print(root)
-  }
-  
+  request_body <- as(root, "character")
   httr_response <- rPOST(url = base_metadata_url,
                          headers = c("SOAPAction"=which_operation,
                                      "Content-Type"="text/xml"),
-                         body = as(root, "character"))
+                         body = request_body)
+  if(verbose){
+    make_verbose_httr_message(httr_response$request$method, 
+                              httr_response$request$url, 
+                              httr_response$request$headers,
+                              request_body)
+  }
   catch_errors(httr_response)
   response_parsed <- content(httr_response, encoding="UTF-8")
   
