@@ -18,6 +18,6 @@ if(getRversion() >= "2.15.1")  utils::globalVariables(c(".",
                                                         "id", 
                                                         "success", 
                                                         "sObjectType",
-                                                        "sobject",
+                                                        "sObject",
                                                         "attributes.type", 
                                                         "sf:type"))
\ No newline at end of file
