@@ -6,9 +6,10 @@
 #' \code{tbl_df}; data can be coerced into a \code{data.frame}
 #' @template object_name
 #' @template external_id_fieldname
-#' @template all_or_none
 #' @template api_type
-#' @param ... Other arguments passed on to \code{\link{sf_bulk_operation}}.
+#' @template control
+#' @param ... arguments passed to \code{\link{sf_control}} or further downstream 
+#' to \code{\link{sf_bulk_operation}}
 #' @template verbose
 #' @return \code{tbl_df} of records with success indicator
 #' @examples
@@ -35,27 +36,53 @@
 sf_upsert <- function(input_data,
                       object_name,
                       external_id_fieldname,
-                      all_or_none = FALSE,
                       api_type = c("SOAP", "REST", "Bulk 1.0", "Bulk 2.0"),
-                      ...,
+                      control = list(...), ...,
                       verbose = FALSE){
   
   api_type <- match.arg(api_type)
-  if(api_type == "REST"){
-    resultset <- sf_upsert_rest(input_data=input_data, object_name=object_name,
-                                external_id_fieldname=external_id_fieldname,
-                                all_or_none=all_or_none, verbose=verbose)
-  } else if(api_type == "SOAP"){
-    resultset <- sf_upsert_soap(input_data=input_data, object_name=object_name,
-                                external_id_fieldname=external_id_fieldname,
-                                all_or_none=all_or_none, verbose=verbose)
+  
+  # determine how to pass along the control args 
+  all_args <- list(...)
+  control_args <- return_matching_controls(control)
+  control_args$api_type <- api_type
+  control_args$operation <- "upsert"
+  if("all_or_none" %in% names(all_args)){
+    # warn then set it in the control list
+    warning(paste0("The `all_or_none` argument has been deprecated.\n", 
+                   "Please pass AllOrNoneHeader argument or use the `sf_control` function."), 
+            call. = FALSE)
+    control_args$AllOrNoneHeader = list(allOrNone = tolower(all_args$all_or_none))
+  }
+  if("AssignmentRuleHeader" %in% names(control_args)){
+    if(!object_name %in% c("Account", "Case", "Lead")){
+      stop("The AssignmentRuleHeader can only be used when creating, updating, or upserting an Account, Case, or Lead")
+    }
+  }
+  
+  if(api_type == "SOAP"){
+    resultset <- sf_upsert_soap(input_data = input_data, 
+                                object_name = object_name,
+                                external_id_fieldname = external_id_fieldname,
+                                control = control_args,
+                                verbose = verbose)
+  } else if(api_type == "REST"){
+    resultset <- sf_upsert_rest(input_data = input_data, 
+                                object_name = object_name,
+                                external_id_fieldname = external_id_fieldname,
+                                control = control_args,
+                                verbose = verbose)
   } else if(api_type == "Bulk 1.0"){
-    resultset <- sf_upsert_bulk_v1(input_data=input_data, object_name=object_name,
-                                   external_id_fieldname=external_id_fieldname,
+    resultset <- sf_upsert_bulk_v1(input_data = input_data, 
+                                   object_name = object_name,
+                                   external_id_fieldname = external_id_fieldname, 
+                                   control = control_args, 
                                    verbose = verbose, ...)
   } else if(api_type == "Bulk 2.0"){
-    resultset <- sf_upsert_bulk_v2(input_data=input_data, object_name=object_name,
-                                   external_id_fieldname=external_id_fieldname,
+    resultset <- sf_upsert_bulk_v2(input_data = input_data, 
+                                   object_name = object_name,
+                                   external_id_fieldname = external_id_fieldname, 
+                                   control = control_args, 
                                    verbose = verbose, ...)
   } else {
     stop("Unknown API type")
@@ -63,7 +90,6 @@ sf_upsert <- function(input_data,
   return(resultset)
 }
 
-
 #' Upsert Records using SOAP API
 #' 
 #' @importFrom readr cols type_convert
@@ -75,26 +101,22 @@ sf_upsert <- function(input_data,
 #' @importFrom utils head
 #' @note This function is meant to be used internally. Only use when debugging.
 #' @keywords internal
-sf_upsert_soap <- function(input_data, object_name, 
+sf_upsert_soap <- function(input_data, 
+                           object_name, 
                            external_id_fieldname,
-                           all_or_none = FALSE,
+                           control, ...,
                            verbose = FALSE){
   
   input_data <- sf_input_data_validation(operation='upsert', input_data)
+  control <- do.call("sf_control", control)
   
   base_soap_url <- make_base_soap_url()
-  if(verbose) {
-    message(base_soap_url)
-  }
-  
   # limit this type of request to only 200 records at a time to prevent 
   # the XML from exceeding a size limit
   batch_size <- 200
   row_num <- nrow(input_data)
   batch_id <- (seq.int(row_num)-1) %/% batch_size
-  if(verbose){
-    message("Submitting data in ", max(batch_id)+1, " Batches")
-  }
+  if(verbose) message("Submitting data in ", max(batch_id)+1, " Batches")
   message_flag <- unique(as.integer(quantile(0:max(batch_id), c(0.25,0.5,0.75,1))))
   
   resultset <- NULL
@@ -106,16 +128,23 @@ sf_upsert_soap <- function(input_data, object_name,
       } 
     }
     batched_data <- input_data[batch_id == batch, , drop=FALSE]  
-    r <- make_soap_xml_skeleton(soap_headers=list(AllorNoneHeader = tolower(all_or_none)))
+    r <- make_soap_xml_skeleton(soap_headers = control)
     xml_dat <- build_soap_xml_from_list(input_data = batched_data,
                                         operation = "upsert",
                                         object_name = object_name,
                                         external_id_fieldname = external_id_fieldname,
                                         root = r)
+    request_body <- as(xml_dat, "character")
     httr_response <- rPOST(url = base_soap_url,
                            headers = c("SOAPAction"="upsert",
                                        "Content-Type"="text/xml"),
-                           body = as(xml_dat, "character"))
+                           body = request_body)
+    if(verbose){
+      make_verbose_httr_message(httr_response$request$method,
+                                httr_response$request$url, 
+                                httr_response$request$headers, 
+                                request_body)
+    }
     catch_errors(httr_response)
     response_parsed <- content(httr_response, encoding="UTF-8")
     this_set <- response_parsed %>%
@@ -129,37 +158,42 @@ sf_upsert_soap <- function(input_data, object_name,
   return(resultset)
 }
 
-
 #' Upsert Records using REST API
 #' 
 #' @importFrom readr cols type_convert
 #' @importFrom dplyr everything as_tibble bind_rows select
-#' @importFrom jsonlite toJSON fromJSON
+#' @importFrom jsonlite toJSON fromJSON prettify
 #' @importFrom stats quantile
 #' @importFrom utils head
 #' @note This function is meant to be used internally. Only use when debugging.
 #' @keywords internal
-sf_upsert_rest <- function(input_data, object_name, 
-                           external_id_fieldname,
-                           all_or_none = FALSE,
+sf_upsert_rest <- function(input_data, 
+                           object_name, 
+                           external_id_fieldname, 
+                           control, ...,
                            verbose = FALSE){
   
   # This resource is available in API version 42.0 and later.
   stopifnot(as.numeric(getOption("salesforcer.api_version")) >= 42.0)
   input_data <- sf_input_data_validation(operation='upsert', input_data)
   
-  composite_batch_url <- make_composite_batch_url()
-  if(verbose) {
-    message(composite_batch_url)
+  control <- do.call("sf_control", control)
+  if("AllOrNoneHeader" %in% names(control)){
+    stop("'All or None' control is not honored when upserting over the REST API. Try using the SOAP API instead.")
+  }
+  request_headers <- c("Accept"="application/json", 
+                       "Content-Type"="application/json")
+  if("AssignmentRuleHeader" %in% names(control)){
+    # take the first list element because it could be useDefaultRule (T/F) or assignmentRuleId
+    request_headers <- c(request_headers, c("Sforce-Auto-Assign" = control$AssignmentRuleHeader[[1]]))
   }
   
+  composite_batch_url <- make_composite_batch_url()
   # limit this type of request to only 25 records, the max for this type of request
   batch_size <- 25
   row_num <- nrow(input_data)
   batch_id <- (seq.int(row_num)-1) %/% batch_size
-  if(verbose){
-    message("Submitting data in ", max(batch_id)+1, " Batches")
-  }
+  if(verbose) message("Submitting data in ", max(batch_id) + 1, " Batches")
   message_flag <- unique(as.integer(quantile(0:max(batch_id), c(0.25,0.5,0.75,1))))
   
   resultset <- NULL
@@ -176,24 +210,29 @@ sf_upsert_rest <- function(input_data, object_name,
       this_id <- batched_data[i, external_id_fieldname]
       temp_batched_data <- batched_data[,-(which(names(batched_data) == external_id_fieldname))]
       temp_batched_data <- temp_batched_data[,!grepl("^ID$|^IDS$", names(temp_batched_data), ignore.case=TRUE)]
-      inner_requests[[i]] <- list(method="PATCH", 
-                                  url=paste0("v", getOption("salesforcer.api_version"), 
-                                             "/sobjects/", object_name, "/", 
-                                             external_id_fieldname, "/", this_id), 
-                                  richInput=as.list(temp_batched_data[i,]))
+      inner_requests[[i]] <- list(method = "PATCH", 
+                                  url = paste0("v", getOption("salesforcer.api_version"), 
+                                               "/sobjects/", object_name, "/", 
+                                               external_id_fieldname, "/", this_id), 
+                                  richInput = as.list(temp_batched_data[i,]))
     }
-    request_body <- list(batchRequests=inner_requests)
+    request_body <- toJSON(list(batchRequests = inner_requests),
+                           auto_unbox = TRUE)
     httr_response <- rPOST(url = composite_batch_url,
-                           headers = c("Accept"="application/json", 
-                                       "Content-Type"="application/json"),
-                           body = toJSON(request_body,
-                                         auto_unbox = TRUE))
+                           headers = request_headers,
+                           body = request_body)
+    if(verbose){
+      make_verbose_httr_message(httr_response$request$method,
+                                httr_response$request$url, 
+                                httr_response$request$headers, 
+                                prettify(request_body))
+    }
     catch_errors(httr_response)
     response_parsed <- content(httr_response, "text", encoding="UTF-8")
     response_parsed <- fromJSON(response_parsed, flatten=TRUE)$results
     response_parsed <- response_parsed %>%
       rename_at(.vars = vars(starts_with("result.")), 
-                .funs = funs(sub("^result\\.", "", .))) %>%
+                .funs = list(~sub("^result\\.", "", .))) %>%
       select(-matches("statusCode")) %>%
       as_tibble()
     resultset <- bind_rows(resultset, response_parsed)
@@ -202,46 +241,47 @@ sf_upsert_rest <- function(input_data, object_name,
     type_convert(col_types = cols())
   return(resultset)
 } 
-  
 
 #' Upsert Records using Bulk 1.0 API
 #' 
 #' @note This function is meant to be used internally. Only use when debugging.
 #' @keywords internal
-sf_upsert_bulk_v1 <- function(input_data, object_name, 
+sf_upsert_bulk_v1 <- function(input_data, 
+                              object_name, 
                               external_id_fieldname,
-                              all_or_none = FALSE,
-                              ...,
+                              control, ...,
                               verbose = FALSE){
-  # allor none?
-  input_data <- sf_input_data_validation(operation="upsert", input_data)
-  resultset <- sf_bulk_operation(input_data=input_data, 
-                                 object_name=object_name,
-                                 external_id_fieldname=external_id_fieldname,
-                                 operation="upsert", 
+  input_data <- sf_input_data_validation(operation = "upsert", input_data)
+  control <- do.call("sf_control", control)
+  resultset <- sf_bulk_operation(input_data = input_data, 
+                                 object_name = object_name,
+                                 external_id_fieldname = external_id_fieldname,
+                                 operation = "upsert", 
                                  api_type = "Bulk 1.0",
-                                 verbose=verbose, ...)
+                                 control = control, ...,
+                                 verbose = verbose)
   return(resultset)
 }
 
-
 #' Upsert Records using Bulk 2.0 API
 #' 
 #' @note This function is meant to be used internally. Only use when debugging.
 #' @keywords internal
-sf_upsert_bulk_v2 <- function(input_data, object_name, 
+sf_upsert_bulk_v2 <- function(input_data, 
+                              object_name, 
                               external_id_fieldname,
-                              all_or_none = FALSE,
-                              ...,
+                              control, ...,
                               verbose = FALSE){
-  # allor none?
-  #The order of records in the response is not guaranteed to match the ordering of records in the original job data.
-  input_data <- sf_input_data_validation(operation='upsert', input_data)
-  resultset <- sf_bulk_operation(input_data=input_data, 
-                                 object_name=object_name,
-                                 external_id_fieldname=external_id_fieldname,
-                                 operation="upsert", 
+  # The order of records in the response is not guaranteed to match the ordering of
+  # records in the original job data.
+  input_data <- sf_input_data_validation(operation = "upsert", input_data)
+  control <- do.call("sf_control", control)
+  resultset <- sf_bulk_operation(input_data = input_data, 
+                                 object_name = object_name,
+                                 external_id_fieldname = external_id_fieldname,
+                                 operation = "upsert", 
                                  api_type = "Bulk 2.0",
-                                 verbose=verbose, ...)
+                                 control = control, ...,
+                                 verbose = verbose)
   return(resultset)
 }
